;;; Funciones para implementacion de los algoritmo A y A*
;;;
;;; A utilizar con q_search.lsp (a generic framework for search)
;;; para implementar dichos algoritmos
;;;
;;; Asignatura: Introduccin a la Inteligencia Artificial
;;;             CSI - ULL
;;; Prctica: 5a (sistemas de produccin)
;;; Profesor: Patricio Garca Bez (patricio@csi.ull.es)
;;; Fecha: 17/5/2002
;;; Comentarios:
;;;           
;;;           Para la correcta utilizacin han de almacenarse
;;;           en las siguientes variables los nombres de las
;;;           funciones especficas a cada problema:
;;;           
;;;           *eval-fn* : 
;;;                       (Definida en q_search.lsp) Funcin que
;;;                       evalua el coste estimado de la solucin
;;;                       mas cercana (f') que pasa por un estado 
;;;                       dado (g + h'). Si se desea implementar 
;;;                       el algoritmo A* esta funcin ha de ser 
;;;                       subestimada.
;;;           
;;;           *coste-acum-fn* :
;;;                       Funcin que calcula el coste acumulado (g)
;;;                       para llegar a un estado dado desde el
;;;                       estado inicial.
;;;           
;;;           *igual-est-fn* :
;;;                       Predicado que compara si dos estados son
;;;                       iguales o no.
;;;
;;;           Ver el correspondiente help para cada una de ellas.
;;;           
;;;           Las variables anteriores son utilizadas por las dos
;;;           funciones genricas que aqu se definen:  
;;;           priority-a-merge-fn y priority-a-dequeue-fn
;;;           y que servirn respectivamente como valores de los 
;;;           parmetros :merge-fn y :dequeue-fn para la funcin
;;;           generic-search definida en q_search.lsp.
;;;           
;;;           Tener en cuenta que en las anteriores funciones no se
;;;           ha impelmentado el tratamiento de estados visitados,
;;;           a parte de aquellos que se encuentran repetidos en 
;;;           la lista frente, siendo responsabilidad del usuario
;;;           su implementacin si as se cree necesaria.
;;;


(defvar *coste-acum-fn* nil
  " - Variable: TIPO funcin

Esta variable guarda la funcin que devuelve el coste acumulado (g) de un estado.")

(defvar *igual-est-fn* nil
  " - Variable: TIPO funcin

Esta variable guarda la funcin que compara dos estado y dice si son iguales.")

(defun crea-nodo (est)
  " - Funcin: CREA-NODO (EST)

Crea nodo tipo (f' . estado) computando la funcin heuristica subdividida f' (*eval-fn*) del estado est."
  (cons (funcall *eval-fn* est) est))
		
(defun igual-y-peor-o-igual (est1 est2)
  (and (funcall *igual-est-fn* est1 est2)
       (>= (funcall *coste-acum-fn* est1)
	   (funcall *coste-acum-fn* est2))))

(defun igual-y-peor (est1 est2)
  (and (funcall *igual-est-fn* est1 est2)
       (> (funcall *coste-acum-fn* est1)
	  (funcall *coste-acum-fn* est2))))

(defun criba-iguales (new-states)
  " - Funcin: CRIBA-IGUALES (NEW-STATE)

Busca si hay elementos de la lista de hijos (new-states) en la lista de frente mediante la funcin almacenada en *igual-est-fn* y si es as criba los que tengan coste acumulado (funcin almacenada en *coste-acum-fn*) mayor de la lista de hijos o de la lista de frente."
  ; Elimina los elementos de new-states que sean iguales a alguno
  ; de la lista frente pero con coste acumulado mayor o igual
  (dolist (el *search-queue*)
    (setf new-states 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor-o-igual (rest est)
					       (rest el)))
		     new-states 
		     :count 1)))
  ; Elimina los elementos de *search-queue* que sean iguales a alguno
  ; de la new-states pero con coste acumulado mayor
  (dolist (el new-states new-states)
    (setf *search-queue* 
	  (delete-if #'(lambda (est) 
			 (igual-y-peor (rest est) (rest el)))
		     *search-queue*
		     :count 1))))

(defun fusion (list1 list2)
  " - Funcin: FUSION (LIST1 LIST2)

Realiza la fusin ordenada segun f' de la lista de hijos (new-states) en lista frente (*search-queue*)."
  (cond
   ((endp list1) list2)
   ((endp list2) list1)
   ((<= (first (first list1))
	(first (first list2)))
    (setf (rest list1) (fusion (rest list1) list2))
    list1)
   (t (setf (rest list2) (fusion list1 (rest list2)))
      list2)))

(defun priority-a-merge-fn (new-states)
  " - Funcin: PRIORITY-A-MERGE-FN (NEW-STATE)

Fusiona new-states a la lista de frente de valores (f' . estado), mantniendola ordenada segun f' y eliminando los peores estados duplicados."
  ; Computa f' de new-states, generando lista de (f' . estado)
  (setf new-states (mapcar #'crea-nodo new-states))
  ; Chequea y criba si los estados hijos ya estan
  ; en la lista de frente de una u otra lista
  (setf new-states (criba-iguales new-states))
  ; Ordena los estados hijos en funcin de f'
  (setf new-states (sort new-states #'< :key #'first))
  ; Realiza fusin ordenada de hijos en lista frente
  (fusion new-states *search-queue*))

(defun priority-a-dequeue-fn ()
  " - Funcin: PRIORITY-A-DEQUEUE-FN ()

Funcin que extrae el primer elemento de la lista de frente creada con  priority-a-merge-fn devolviendo un estado."
  (rest (pop *search-queue*)))

